@@ -44,7 +44,7 @@ from .utils import JSessionAuth
 
 GEN_MODULES = {}
 
-__version__ = '0.3.28'
+__version__ = '0.4.0'
 __all__ = ['connect', 'exceptions']
 
 
@@ -486,9 +486,17 @@ def connect(server, user=None, password=None, verify=True, netrc_file=None, debu
     # Start out with any accidental auth, fill token once retrieved
     requests_session.auth = JSessionAuth()
 
+    # Remove port and add .local to the domain in case there is no . in host
+    # See issue #5388 in psf/requests for more info 
+    domain = parse.urlparse(server).netloc
+    if ":" in domain:
+        domain = domain.split(":")[0]
+    if "." not in domain:
+        domain = "{domain}.local".format(domain=domain)
+
     if jsession is not None:
         cookie = requests.cookies.create_cookie(
-            domain=parse.urlparse(server).netloc,
+            domain=domain,
             name='JSESSIONID',
             value=jsession,
         )
@@ -531,7 +539,7 @@ def connect(server, user=None, password=None, verify=True, netrc_file=None, debu
         logger.debug("Requests session cookies: {}".format(requests_session.cookies))
     else:
         logger.debug("Set JSESSION_TOKEN to: {}".format(jsession))
-        if requests_session.cookies['JSESSIONID'] != jsession:
+        if requests_session.cookies.get('JSESSIONID', domain=domain) != jsession:
             message = 'Failed to login by re-using existing jsession, the session is probably close on the server!'
             logger.error(message)
             raise exceptions.XNATLoginFailedError(message)
@@ -562,10 +570,7 @@ def connect(server, user=None, password=None, verify=True, netrc_file=None, debu
         xnat_session = SessionType(server=server, logger=logger,
                                    interface=requests_session, debug=debug,
                                    original_uri=original_uri, logged_in_user=logged_in_user,
-                                   default_timeout=default_timeout)
-
-        # Store JSESSION for completeness
-        xnat_session.JSESSION = jsession_token
+                                   default_timeout=default_timeout, jsession=jsession_token)
 
         # Parse data model and create classes
         if not no_parse_model:
